using System;
using Microsoft.SPOT;
using GHIElectronics.NETMF.System;

namespace HighAltitude
{
    class ImuCalculator
    {
        /// <summary>AirSpeed in m/s</summary>
        public float InpAirSpeed = 0;

        /// <summary>Acceleration in m/s</summary>
        public Vector3D InpAcc = new Vector3D();

        /// <summary>Angular rate in deg/s</summary>
        public Vector3D InpGyr = new Vector3D();

        public Vector3D InpMag = new Vector3D();

        /// <summary>Ascent/descent angle</summary>
        public float Pitch = 0;
        /// <summary>Roll angle</summary>
        public float Roll = 0;
        /// <summary>Attitude angle</summary>
        public float Yaw = 0;

        private float CorrAccZ = 0;
        //private DateTime lastTimePoint = new DateTime();

        private Quaternion rotaQuat = new Quaternion();
        private Vector3D VertVector = new Vector3D(0, 0, 1);
        private Vector3D EastVector = new Vector3D(1, 0, 0);

        //float errx = 0; //Roll-Pitch-Yaw
        //float erry = 0;
        //float errz = 0;

        public void Calculate(float dt)
        {
            float dx, dy, dz;
            

            //Correcting centrifugal acceleration
            //CorrAccZ = InpAcc.Z - InpAirSpeed * InpGyr.X / degPerPi; //Check sign!

            if (dt < 0.5 && dt > 0)
            {
                dx = InpGyr.X * 1.2f * dt;
                dy = InpGyr.Y * 1.2f * dt;
                dz = InpGyr.Z * 1.2f * dt;

                //Rotation dt = 1.22 ms
                rotaQuat.XAxisTurn(dx);
                rotaQuat.YAxisTurn(dy);
                rotaQuat.ZAxisTurn(dz);


                //Angle calculations dt = 3.35 ms
                var absInpAcc = rotaQuat.RotateVector(InpAcc);
                var absInpMag = rotaQuat.RotateVector(InpMag) * VertVector;
                var angleXY = Vector3D.VectorAngle(absInpAcc, VertVector);
                var angleZ = Vector3D.VectorAngle(absInpMag, EastVector);

                //if (angleXY > 10) angleXY = 10;

                //Drift Angle dt = 3.35 ms
                var AccQuat = new Quaternion(VertVector * absInpAcc, -angleXY / 120);
                var MagQuat = new Quaternion(EastVector * absInpMag, -angleZ / 40);

                rotaQuat = AccQuat * MagQuat * rotaQuat; //MagQuat

                //lastTimePoint = DateTime.UtcNow;

                //1 ms
                var eulerAngles = rotaQuat.GetEulerAngles();
                Pitch = eulerAngles.X;
                Roll = eulerAngles.Y;
                Yaw = eulerAngles.Z;


                //Debug.Print(absInpAcc.X + ";   " + absInpAcc.Y + ";   " + absInpAcc.Z + ";");
                //Debug.Print("X/Y/Z: " + Pitch + "    /    " + Roll + "   /   " + Yaw + " dt " + dt);

               // Debug.Print(InpGyr.X + "; " + InpGyr.Y + "; " + InpGyr.Z + "; " + angleXY);
            }
        }
    }
}
